# +---------------------------------------------------------------------+
# | (c) 2025 Mario Sieg <mario.sieg.64@gmail.com>                       |
# | Licensed under the Apache License, Version 2.0                      |
# |                                                                     |
# | Website : https://mariosieg.com                                     |
# | GitHub  : https://github.com/MarioSieg                              |
# | License : https://www.apache.org/licenses/LICENSE-2.0               |
# +---------------------------------------------------------------------+

from __future__ import annotations

import weakref
from contextlib import ContextDecorator
from dataclasses import dataclass
from types import TracebackType

from ._bootstrap import _FFI, _C


@dataclass
class NativeErrorInfo:
    status_code: int
    message: str
    file: str
    line: int
    col: int
    func: str


_ctx = _C.mag_ctx_create(bytes('cpu', 'utf-8'))
_ctx_finalizer = weakref.finalize(_ctx, _C.mag_ctx_destroy, _ctx, True)


def native_ptr() -> _FFI.CData:
    return _ctx


def last_error_code() -> int:
    return _C.mag_ctx_get_last_error_code(_ctx)


def has_last_error() -> bool:
    return _C.mag_ctx_has_error(_ctx)


def clear_last_error() -> None:
    _C.mag_ctx_clear_last_error(_ctx)


def last_error_str() -> str:
    return _C.mag_status_get_name(_C.mag_ctx_get_last_error_code(_ctx)).decode('utf-8')


def take_last_error() -> NativeErrorInfo | None:
    if not has_last_error():
        return None
    native_err: _FFI.CData = _FFI.new('mag_error_t*')
    _C.mag_ctx_take_last_error(_ctx, native_err)
    status_code: int = native_err.code
    message: str = _FFI.string(native_err.message).decode('utf-8')
    file: str = _FFI.string(native_err.file).decode('utf-8')
    line: int = native_err.line
    col: int = native_err.col
    func: str = _FFI.string(native_err.func).decode('utf-8')
    return NativeErrorInfo(status_code, message, file, line, col, func)


def start_grad_recorder() -> None:
    _C.mag_ctx_grad_recorder_start(_ctx)


def stop_grad_recorder() -> None:
    _C.mag_ctx_grad_recorder_stop(_ctx)


def is_grad_recording() -> bool:
    return _C.mag_ctx_grad_recorder_is_running(_ctx)


def manual_seed(seed: int) -> None:
    if seed < 0:
        seed += 0xFFFFFFFFFFFFFFFF
    seed &= 0xFFFFFFFFFFFFFFFF
    _C.mag_ctx_manual_seed(_ctx, seed)


def compute_device_name() -> str:
    return _FFI.string(_C.mag_ctx_get_compute_device_name(_ctx)).decode('utf-8')


def os_name() -> str:
    return _FFI.string(_C.mag_ctx_get_os_name(_ctx)).decode('utf-8')


def cpu_name() -> str:
    return _FFI.string(_C.mag_ctx_get_cpu_name(_ctx)).decode('utf-8')


def cpu_virtual_cores() -> int:
    return _C.mag_ctx_get_cpu_virtual_cores(_ctx)


def cpu_physical_cores() -> int:
    return _C.mag_ctx_get_cpu_physical_cores(_ctx)


def cpu_sockets() -> int:
    return _C.mag_ctx_get_cpu_sockets(_ctx)


def physical_memory_total() -> int:
    return _C.mag_ctx_get_physical_memory_total(_ctx)


def physical_memory_free() -> int:
    return _C.mag_ctx_get_physical_memory_free(_ctx)


def physical_memory_used() -> int:
    return abs(physical_memory_total() - physical_memory_free())


def is_numa_system() -> bool:
    return _C.mag_ctx_is_numa_system(_ctx)


class no_grad(ContextDecorator):
    """Disables gradient recording within a function or block."""

    def __enter__(self) -> None:
        """Disable gradient tracking by stopping the active context's recorder."""
        stop_grad_recorder()

    def __exit__(self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None) -> None:
        """Re-enable gradient tracking when exiting the context."""
        start_grad_recorder()



# (c) 2025 Mario Sieg. <mario.sieg.64@gmail.com>

import torch

from ..common import *


def _make_tensor(shape, dtype):
    if dtype == boolean:
        return Tensor.bernoulli(shape)
    return Tensor.uniform(shape, dtype=dtype)


def _rand_shape(max_rank=4, max_size=6):
    rank = random.randint(1, max_rank)
    return tuple(random.randint(1, max_size) for _ in range(rank))


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_clone(dtype: DataType) -> None:
    def func(shape: tuple[int, ...]) -> None:
        x = _make_tensor(shape, dtype)
        a = x
        b = a.clone()
        assert a.shape == b.shape
        assert a.strides == b.strides
        assert a.numel == b.numel
        assert a.rank == b.rank
        assert a.is_contiguous == b.is_contiguous
        assert a.tolist() == b.tolist()

    for_all_shapes(func)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_gather(dtype: DataType) -> None:
    def func(shape: tuple[int, ...]) -> None:
        x = _make_tensor(shape, dtype)
        torch_x = totorch(x)
        rank = len(shape)
        for dim in range(rank):
            index_torch = torch.randint(0, shape[dim], size=shape, dtype=torch.int64)
            index_own = Tensor.of(index_torch.tolist(), dtype=int64)
            out_own = x.gather(dim, index_own)
            out_torch = torch.gather(torch_x, dim, index_torch)
            torch.testing.assert_close(totorch(out_own), out_torch)

    for_all_shapes(func)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_cat_random(dtype):
    base_shape = _rand_shape()
    rank = len(base_shape)
    dim = random.randint(0, rank - 1)
    k = random.randint(2, 5)
    others = list(base_shape)
    sizes = [random.randint(2, 5) for _ in range(k)]
    shapes = [tuple(others[:dim] + [s] + others[dim + 1 :]) for s in sizes]
    xs = [_make_tensor(s, dtype) for s in shapes]
    ts = [totorch(x) for x in xs]
    y = Tensor.cat(xs, dim=dim)
    yt = torch.cat(ts, dim=dim)
    torch.testing.assert_close(totorch(y), yt)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_cat_negative_dim(dtype):
    base_shape = _rand_shape()
    dim = -1
    k = random.randint(2, 4)
    others = list(base_shape)
    sizes = [random.randint(1, 4) for _ in range(k)]
    shapes = [tuple(others[:-1] + [s]) for s in sizes]
    xs = [_make_tensor(s, dtype) for s in shapes]
    ts = [totorch(x) for x in xs]

    y = Tensor.cat(xs, dim=dim)
    yt = torch.cat(ts, dim=dim)

    torch.testing.assert_close(totorch(y), yt)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_cat_single_tensor_noop(dtype):
    shape = _rand_shape()
    x = _make_tensor(shape, dtype)
    y = Tensor.cat([x], dim=0)
    torch.testing.assert_close(totorch(y), totorch(x))


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_cat_shape_mismatch_raises(dtype):
    rank = random.randint(2, 4)
    shape_a = [random.randint(2, 5) for _ in range(rank)]
    shape_b = shape_a[:]
    dim = random.randint(0, rank - 1)

    other = (dim + 1) % rank
    shape_b[other] = shape_a[other] + 1

    a = _make_tensor(tuple(shape_a), dtype)
    b = _make_tensor(tuple(shape_b), dtype)

    with pytest.raises(AssertionError):
        _ = Tensor.cat([a, b], dim=dim)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_cat_matches_torch_many_chunks(dtype):
    base_shape = _rand_shape()
    rank = len(base_shape)
    dim = random.randint(0, rank - 1)

    others = list(base_shape)
    k = 10
    sizes = [random.randint(2, 5) for _ in range(k)]
    shapes = [tuple(others[:dim] + [s] + others[dim + 1 :]) for s in sizes]

    xs = [_make_tensor(s, dtype) for s in shapes]
    ts = [totorch(x) for x in xs]

    y = Tensor.cat(xs, dim=dim)
    yt = torch.cat(ts, dim=dim)
    torch.testing.assert_close(totorch(y), yt)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_split_cat_roundtrip_random(dtype):
    for _ in range(1000):
        shape = _rand_shape()
        x = _make_tensor(shape, dtype)
        rank = len(shape)
        dim = random.randint(0, rank - 1)
        split_size = random.randint(1, shape[dim])
        parts = x.split(split_size, dim)
        assert len(parts) >= 1
        y = Tensor.cat(parts, dim=dim)
        assert x.tolist() == y.tolist()


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_split(dtype: DataType) -> None:
    def func(shape: tuple[int, ...]) -> None:
        x = _make_tensor(shape, dtype)
        if len(shape) <= 1:
            return
        dim = random.randint(0, len(shape) - 1)
        split_size = random.randint(1, shape[dim])
        a = x.split(split_size, dim)
        b = totorch(x).split(split_size, dim)
        assert len(a) == len(b)
        for i in range(len(a)):
            torch.testing.assert_close(totorch(a[i]), b[i])

    for_all_shapes(func)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_tolist(dtype: DataType) -> None:
    def func(shape: tuple[int, ...]) -> None:
        x = _make_tensor(shape, dtype)
        assert x.tolist() == totorch(x).tolist()

    for_all_shapes(func)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_transpose(dtype: DataType) -> None:
    def func(shape: tuple[int, ...]) -> None:
        if len(shape) <= 1:
            return
        if dtype == boolean:
            x = Tensor.bernoulli(shape)
        else:
            x = Tensor.uniform(shape, dtype=dtype)
        sample = lambda: random.randint(-len(shape) + 1, len(shape) - 1)
        dim1: int = sample()
        dim2: int = dim1
        while dim2 == dim1:  # Reject equal transposition axes
            dim2 = sample()
        a = totorch(x.transpose(dim1, dim2))
        b = totorch(x).transpose(dim1, dim2)
        if not torch.allclose(a, b):
            print('M=' + str(a))
            print('T=' + str(b))
            print(f'axes: {dim1} {dim2}')
        torch.testing.assert_close(a, b)

    for_all_shapes(func)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_view(dtype: DataType) -> None:
    def func(shape: tuple[int, ...]) -> None:
        x = _make_tensor(shape, dtype)
        shape = list(shape)
        random.shuffle(shape)  # Shuffle view shape
        shape = tuple(shape)
        y = x.view(*shape)
        torch.testing.assert_close(totorch(y), totorch(x).view(shape))

    for_all_shapes(func)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_view_infer_axis(dtype: DataType) -> None:
    def func(shape: tuple[int, ...]) -> None:
        x = _make_tensor(shape, dtype)
        shape = list(shape)
        random.shuffle(shape)  # Shuffle view shape
        shape[random.randint(0, len(shape) - 1)] = -1  # Set inferred axis randomly
        shape = tuple(shape)
        y = x.view(*shape)
        torch.testing.assert_close(totorch(y), totorch(x).view(shape))

    for_all_shapes(func)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_reshape(dtype: DataType) -> None:
    def func(shape: tuple[int, ...]) -> None:
        x = _make_tensor(shape, dtype)
        shape = list(shape)
        random.shuffle(shape)  # Shuffle reshape shape
        shape = tuple(shape)
        y = x.T.reshape(*shape)
        torch.testing.assert_close(totorch(y), totorch(y).reshape(shape))

    for_all_shapes(func)


@pytest.mark.parametrize('dtype', [float16, float32, boolean, int32])
def test_tensor_reshape_infer_axis(dtype: DataType) -> None:
    def func(shape: tuple[int, ...]) -> None:
        x = _make_tensor(shape, dtype)
        shape = list(shape)
        random.shuffle(shape)  # Shuffle reshape shape
        shape[random.randint(0, len(shape) - 1)] = -1  # Set inferred axis randomly
        shape = tuple(shape)
        y = x.T.reshape(*shape)
        torch.testing.assert_close(totorch(y), totorch(y).reshape(shape))

    for_all_shapes(func)


def test_tensor_permute() -> None:
    a = Tensor.full(2, 3, fill_value=1)
    b = a.permute((1, 0))
    assert a.shape == (2, 3)
    assert b.shape == (3, 2)
    assert a.numel == 6
    assert b.numel == 6
    assert a.rank == 2
    assert b.rank == 2
    assert a.tolist() == [[1, 1, 1], [1, 1, 1]]
    assert b.tolist() == [[1, 1], [1, 1], [1, 1]]
    assert a.is_contiguous
    assert not b.is_contiguous



# +---------------------------------------------------------------------+
# | (c) 2025 Mario Sieg <mario.sieg.64@gmail.com>                       |
# | Licensed under the Apache License, Version 2.0                      |
# |                                                                     |
# | Website : https://mariosieg.com                                     |
# | GitHub  : https://github.com/MarioSieg                              |
# | License : https://www.apache.org/licenses/LICENSE-2.0               |
# +---------------------------------------------------------------------+

from __future__ import annotations
from dataclasses import dataclass

from ._bootstrap import _FFI, _C

FLOATING_POINT_DTYPES: set[DataType] = set()  # Includes all floating-point types.
INTEGRAL_DTYPES: set[DataType] = set()  # Includes all integral types (integers + boolean).
INTEGER_DTYPES: set[DataType] = set()  # Include all integer types (integers - boolean).
NUMERIC_DTYPES: set[DataType] = set()  # Include all numeric dtypes (floating point + integers - boolean)


@dataclass(frozen=True)
class DataType:
    enum_value: int
    size: int
    name: str
    native_type: str | None
    fill_fn: _FFI.CData

    @property
    def is_floating_point(self) -> bool:
        return self in FLOATING_POINT_DTYPES

    @property
    def is_integral(self) -> bool:
        return self in INTEGRAL_DTYPES

    @property
    def is_integer(self) -> bool:
        return self in INTEGER_DTYPES

    @property
    def is_numeric(self) -> bool:
        return self in NUMERIC_DTYPES

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name


float32: DataType = DataType(_C.MAG_DTYPE_FLOAT32, 4, 'float32', 'float', _C.mag_copy_float_)
float16: DataType = DataType(_C.MAG_DTYPE_FLOAT16, 2, 'float16', None, _C.mag_copy_float_)
boolean: DataType = DataType(_C.MAG_DTYPE_BOOLEAN, 1, 'uint8', 'uint8_t', _C.mag_copy_raw_)
uint8: DataType = DataType(_C.MAG_DTYPE_UINT8, 1, 'uint8', 'uint8_t', _C.mag_copy_raw_)
int8: DataType = DataType(_C.MAG_DTYPE_INT8, 1, 'int8', 'int8_t', _C.mag_copy_raw_)
uint16: DataType = DataType(_C.MAG_DTYPE_UINT16, 2, 'uint16', 'uint16_t', _C.mag_copy_raw_)
int16: DataType = DataType(_C.MAG_DTYPE_INT16, 2, 'int16', 'int16_t', _C.mag_copy_raw_)
uint32: DataType = DataType(_C.MAG_DTYPE_UINT32, 4, 'uint32', 'uint32_t', _C.mag_copy_raw_)
int32: DataType = DataType(_C.MAG_DTYPE_INT32, 4, 'int32', 'int32_t', _C.mag_copy_raw_)
uint64: DataType = DataType(_C.MAG_DTYPE_UINT64, 8, 'uint64', 'uint64_t', _C.mag_copy_raw_)
int64: DataType = DataType(_C.MAG_DTYPE_INT64, 8, 'int64', 'int64_t', _C.mag_copy_raw_)

DTYPE_ENUM_MAP: dict[int, DataType] = {
    float32.enum_value: float32,
    float16.enum_value: float16,
    boolean.enum_value: boolean,
    uint8.enum_value: uint8,
    int8.enum_value: int8,
    uint16.enum_value: uint16,
    int16.enum_value: int16,
    uint32.enum_value: uint32,
    int32.enum_value: int32,
    uint64.enum_value: uint64,
    int64.enum_value: int64,
}
FLOATING_POINT_DTYPES = {float32, float16}
INTEGER_DTYPES = {uint8, int8, uint16, int16, uint32, int32, uint64, int64}
INTEGRAL_DTYPES = INTEGER_DTYPES | {boolean}
NUMERIC_DTYPES = FLOATING_POINT_DTYPES | INTEGER_DTYPES



from __future__ import annotations

from magnetron import Tensor, nn, optim, dtype
from matplotlib import pyplot as plt

EPOCHS: int = 100
LR: float = 1e-1

N = 100
x = -1.0 + Tensor.arange(0, N, 1, dtype=dtype.float32).reshape(N, 1) * (2.0 / (N - 1))
noise = Tensor.normal((N, 1), mean=0.0, std=0.2)
y = 3.0 * x + 0.5 + noise

# Create model, loss function, and optimizer
model = nn.Sequential(nn.Linear(1, 1))
criterion = nn.MSELoss()
opt = optim.SGD(model.parameters(), lr=LR)

# Train the model
losses = []
for epoch in range(EPOCHS):
    pred = model(x)
    loss = criterion(pred, y)
    loss.backward()
    opt.step()
    opt.zero_grad()
    if epoch % 10 == 0:
        print(f'Epoch {epoch}, Loss: {loss.item()}')
    losses.append(loss.item())

y_line = model(x)  # Fitted line

# Plot results
plt.subplot(1, 2, 1)
plt.scatter([x[i].item() for i in range(x.numel)], [y[i].item() for i in range(y.numel)], s=16, label='Data')
plt.plot([x[i].item() for i in range(x.numel)], [y_line[i].item() for i in range(y_line.numel)], c='r', label='Fit')
plt.legend()
plt.title('Linear Regression Fit')

plt.subplot(1, 2, 2)
plt.plot(losses)
plt.title('Loss')
plt.xlabel('Epoch')
plt.tight_layout()
plt.show()



from __future__ import annotations

import codecs
import math
import time
import argparse

from magnetron import Tensor, no_grad, context, dtype
import magnetron.nn as nn
import tiktoken
from collections.abc import Iterator
from dataclasses import dataclass
from magnetron._bootstrap import _FFI, _C
from rich.console import Console

console = Console()

tok = tiktoken.get_encoding('gpt2')

EOS: str = '<|endoftext|>'
encode = lambda x: tok.encode(x, allowed_special={EOS})
decode = lambda x: tok.decode(x)
EOS_ID: int = encode(EOS)[0]


@dataclass
class GPT2HyperParams:
    block_size: int = 1024
    vocab_size: int = 50257
    n_layer: int = 12
    n_head: int = 12
    n_embd: int = 768
    bias: bool = True


class CausalSelfAttention(nn.Module):
    def __init__(self, config: GPT2HyperParams) -> None:
        super().__init__()
        assert config.n_embd % config.n_head == 0
        self.c_attn = nn.Linear(config.n_embd, 3 * config.n_embd, bias=config.bias)
        self.c_proj = nn.Linear(config.n_embd, config.n_embd, bias=config.bias)
        self.n_head = config.n_head
        self.n_embd = config.n_embd
        self.register_buffer(
            'bias',
            Tensor.ones(config.block_size, config.block_size).tril().view(1, 1, config.block_size, config.block_size),
        )

    def forward(self, x: Tensor, kv: tuple[Tensor, Tensor] | None = None) -> tuple[Tensor, tuple[Tensor, Tensor] | None]:
        B, T, C = x.shape
        q, k, v = self.c_attn(x).split(self.n_embd, dim=2)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        if kv is None:
            K, V = k, v
            att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(C // self.n_head))
            att = att.masked_fill(self.bias[:, :, :T, :T] == 0, float('-inf'))
        else:
            K, V = kv
            K = Tensor.cat([K, k], dim=2)
            V = Tensor.cat([V, v], dim=2)
            att = (q @ K.transpose(-2, -1)) * (1.0 / math.sqrt(C // self.n_head))
        y = (att.softmax(dim=-1) @ V).transpose(1, 2).reshape(B, T, C)
        return self.c_proj(y), (K, V)


class MLP(nn.Module):
    def __init__(self, config: GPT2HyperParams) -> None:
        super().__init__()
        self.c_fc = nn.Linear(config.n_embd, 4 * config.n_embd, bias=config.bias)
        self.gelu = nn.GeLU()
        self.c_proj = nn.Linear(4 * config.n_embd, config.n_embd, bias=config.bias)

    def forward(self, x: Tensor) -> Tensor:
        x = self.c_fc(x)
        x = self.gelu(x)
        x = self.c_proj(x)
        return x


class Block(nn.Module):
    def __init__(self, config: GPT2HyperParams) -> None:
        super().__init__()
        self.ln_1 = nn.LayerNorm(config.n_embd, bias=config.bias)
        self.attn = CausalSelfAttention(config)
        self.ln_2 = nn.LayerNorm(config.n_embd, bias=config.bias)
        self.mlp = MLP(config)

    def forward(self, x: Tensor, kv: list[tuple[Tensor, Tensor]] | None = None) -> tuple[Tensor, tuple[Tensor, Tensor]]:
        a, kv = self.attn(self.ln_1(x), kv)
        x = x + a
        x = x + self.mlp(self.ln_2(x))
        return x, kv


class GPT2(nn.Module):
    def __init__(self, config: GPT2HyperParams) -> None:
        super().__init__()
        assert config.vocab_size is not None
        assert config.block_size is not None
        self.config = config
        self.transformer = nn.ModuleDict(
            dict(
                wte=nn.Embedding(config.vocab_size, config.n_embd),
                wpe=nn.Embedding(config.block_size, config.n_embd),
                h=nn.ModuleList([Block(config) for _ in range(config.n_layer)]),
                ln_f=nn.LayerNorm(config.n_embd, bias=config.bias),
            )
        )
        self.lm_head = nn.Linear(config.n_embd, config.vocab_size, bias=False)
        self.transformer.wte.weight = self.lm_head.weight
        console.print(f'Parameter count: {self.get_num_params(False) // 1e6}M', style='dim')

    @classmethod
    @no_grad()
    def from_pretrained(cls, model_type: str = 'gpt2') -> GPT2:
        assert model_type in {'gpt2', 'gpt2-medium', 'gpt2-large', 'gpt2-xl'}
        from transformers import GPT2LMHeadModel

        cfg = {
            'gpt2': dict(n_layer=12, n_head=12, n_embd=768),
            'gpt2-medium': dict(n_layer=24, n_head=16, n_embd=1024),
            'gpt2-large': dict(n_layer=36, n_head=20, n_embd=1280),
            'gpt2-xl': dict(n_layer=48, n_head=25, n_embd=1600),
        }[model_type]
        cfg['vocab_size'] = 50257
        cfg['block_size'] = 1024
        cfg['bias'] = True
        cfg = GPT2HyperParams(**cfg)
        console.print(f'Loading {model_type} with config: {cfg}', style='dim')
        model = cls(cfg)
        sd = model.state_dict()
        sd_keys = sd.keys()
        sd_keys = [k for k in sd_keys if not k.endswith('.attn.bias')]
        model_hf = GPT2LMHeadModel.from_pretrained(model_type)
        sd_hf = model_hf.state_dict()
        sd_keys_hf = sd_hf.keys()
        sd_keys_hf = [k for k in sd_keys_hf if not k.endswith('.attn.masked_bias')]
        sd_keys_hf = [k for k in sd_keys_hf if not k.endswith('.attn.bias')]
        transposed = ['attn.c_attn.weight', 'attn.c_proj.weight', 'mlp.c_fc.weight', 'mlp.c_proj.weight']
        assert len(sd_keys_hf) == len(sd_keys), f'mismatched keys: {len(sd_keys_hf)} != {len(sd_keys)}'

        def copy(r: Tensor, x: torch.Tensor) -> None:  # TODO
            assert x.is_contiguous and r.is_contiguous
            assert r.shape == x.shape, f'Shape mismatch: {r.shape} != {x.shape}'
            assert r.is_contiguous and x.is_contiguous, 'Both tensors must be contiguous for copy operation'
            bytes = x.numel() * x.element_size()
            _C.mag_copy_raw_(r._ptr, _FFI.cast('void*', x.data_ptr()), bytes)

        for k in sd_keys_hf:
            if any(k.endswith(w) for w in transposed):
                assert sd_hf[k].shape[::-1] == sd[k].shape
                with no_grad():
                    copy(sd[k], sd_hf[k].T.contiguous())
            else:
                assert sd_hf[k].shape == sd[k].shape
                with no_grad():
                    copy(sd[k], sd_hf[k].contiguous())

        return model

    def get_num_params(self, non_embedding: bool = False) -> int:
        n_params = sum(p.x.numel for p in self.parameters())
        if non_embedding:
            n_params -= self.transformer.wpe.weight.x.numel
        return n_params

    def forward(self, idx: Tensor, prev_kv: list[tuple[Tensor, Tensor]] | None = None) -> tuple[Tensor, list[tuple[Tensor, Tensor]] | None]:
        b, t = idx.shape
        past_len = 0 if prev_kv is None else prev_kv[0][0].shape[2]
        pos = Tensor.arange(past_len, past_len + t, dtype=dtype.int64)
        x = self.transformer.wte(idx) + self.transformer.wpe(pos)

        new_kv = []
        if prev_kv is None:
            for blk in self.transformer.h:
                x, kv = blk(x, None)
                new_kv.append(kv)
        else:
            for blk, kv_in in zip(self.transformer.h, prev_kv):
                x, kv = blk(x, kv_in)
                new_kv.append(kv)

        x = self.transformer.ln_f(x)
        logits = self.lm_head(x[:, [-1], :])
        return logits, new_kv

    @no_grad()
    def generate(self, prompt: str, max_tokens: int, temp: float = 1.0) -> str:
        tokens = encode(prompt)
        idx = Tensor.of(tokens[-self.config.block_size :] if len(tokens) > self.config.block_size else tokens, dtype=dtype.int64)[None, ...]
        logits, kv = self(idx, prev_kv=None)
        for _ in range(max_tokens):
            probs = (logits[:, -1, :] / temp).softmax(dim=-1)
            next_id = probs.multinomial(num_samples=1).item()
            tokens.append(next_id)
            idx = Tensor.of([next_id], dtype=dtype.int64)[None, ...]
            logits, kv = self(idx, prev_kv=kv)
        return decode(tokens)

    @no_grad()
    def generate_stream(self, prompt: str, max_tokens: int, temp: float = 1.0) -> Iterator[str]:
        tokens = encode(prompt)
        dec = codecs.getincrementaldecoder('utf-8')()
        start = time.perf_counter()
        n = 0
        idx = Tensor.of(tokens[-self.config.block_size :] if len(tokens) > self.config.block_size else tokens, dtype=dtype.int64)[None, ...]
        logits, kv = self(idx, prev_kv=None)
        for _ in range(max_tokens):
            probs = (logits[:, -1, :] / temp).softmax(dim=-1)
            next_id = probs.multinomial(num_samples=1).item()
            tokens.append(next_id)
            n += 1
            delta = dec.decode(tok.decode_single_token_bytes(next_id))
            if delta:
                yield delta
            idx = Tensor.of([next_id], dtype=dtype.int64)[None, ...]
            logits, kv = self(idx, prev_kv=kv)
        tail = dec.decode(b'', final=True)
        if tail:
            yield tail
        elapsed = time.perf_counter() - start
        if n > 0:
            console.print(f'\nTokens/s: {n / elapsed:.2f}, {n} tokens in {elapsed:.3f}s', style='dim')


def _main() -> None:
    args = argparse.ArgumentParser(description='Run GPT-2 model inference')
    args.add_argument('prompt', type=str, help='Prompt to start generation')
    args.add_argument('--model', type=str, default='gpt2', help='Model type (gpt2, gpt2-medium, gpt2-large, gpt2-xl)')
    args.add_argument('--max_tokens', type=int, default=128, help='Maximum number of new tokens to generate')
    args.add_argument('--temp', type=float, default=0.6, help='Temperature for sampling')
    args.add_argument('--no-stream', action='store_true', help='Disable streaming output')
    args.add_argument('--seed', type=int, default=3407, help='Random seed for reproducibility')
    args = args.parse_args()

    context.stop_grad_recorder()
    context.manual_seed(args.seed)

    model = GPT2.from_pretrained(args.model)
    puts = lambda s: console.print(s, style='bold white', end='')
    puts(args.prompt)
    if not args.no_stream:
        for chunk in model.generate_stream(args.prompt, max_tokens=args.max_tokens, temp=args.temp):
            puts(chunk)
    else:
        puts(model.generate(args.prompt, max_tokens=args.max_tokens, temp=args.temp))


if __name__ == '__main__':
    _main()



# (c) 2025 Mario 'Neo' Sieg. <mario.sieg.64@gmail.com>

from __future__ import annotations

from ..common import *

_UNARY_OPS: tuple[str, ...] = (
    'not',
    'abs',
    'neg',
    'log',
    'log10',
    'log1p',
    'log2',
    'sqr',
    'sqrt',
    'sin',
    'asin',
    'sinh',
    'asinh',
    'cos',
    'acos',
    'cosh',
    'acosh',
    'tan',
    'atan',
    'tanh',
    'atanh',
    'step',
    'exp',
    'expm1',
    'exp2',
    'floor',
    'ceil',
    'round',
    'trunc',
    'softmax',
    'sigmoid',
    'hardsigmoid',
    'silu',
    'tanh',
    'gelu',
    'tril',
    'triu',
)


def unary_op(
    dtype: DataType,
    mag_callback: Callable[[Tensor | torch.Tensor], Tensor | torch.Tensor],
    torch_callback: Callable[[Tensor | torch.Tensor], Tensor | torch.Tensor],
) -> None:
    def test(shape: tuple[int, ...]) -> None:
        x = Tensor.bernoulli(shape) if dtype == boolean else Tensor.uniform(shape, dtype=dtype)
        r = mag_callback(x.clone())
        torch.testing.assert_close(totorch(r), torch_callback(totorch(x)).squeeze())

    for_all_shapes(test)


@pytest.mark.parametrize('dtype', FLOATING_POINT_DTYPES)
@pytest.mark.parametrize('op', _UNARY_OPS)
def test_unary_operators(op: str, dtype: DataType) -> None:
    unary_op(dtype, lambda x: getattr(x, op)(), lambda x: getattr(torch, op)(x) if hasattr(torch, op) else getattr(torch.functional.F, op)(x))
    unary_op(dtype, lambda x: getattr(x, op + '_')(), lambda x: getattr(torch, op)(x) if hasattr(torch, op) else getattr(torch.functional.F, op)(x))



# +---------------------------------------------------------------------+
# | (c) 2025 Mario Sieg <mario.sieg.64@gmail.com>                       |
# | Licensed under the Apache License, Version 2.0                      |
# |                                                                     |
# | Website : https://mariosieg.com                                     |
# | GitHub  : https://github.com/MarioSieg                              |
# | License : https://www.apache.org/licenses/LICENSE-2.0               |
# +---------------------------------------------------------------------+

from __future__ import annotations

from . import context
from ._bootstrap import _C, _FFI


class MagnetronError(RuntimeError):
    def __init__(self, message: str, native_info: context.NativeErrorInfo | None) -> None:
        super().__init__(message)
        self.native_info = native_info


def _handle_errc(status: int) -> None:
    if status == _C.MAG_STATUS_OK:
        return
    info: context.NativeErrorInfo | None = context.take_last_error()
    ercc_name: str = _FFI.string(_C.mag_status_get_name(status)).decode('utf-8')
    msg = f'Magnetron C runtime error: 0x{status:X} ({ercc_name})\n'
    if info is not None:
        msg += f'Triggered at {info.file}:{info.line}\n\n'
        msg += f'{info.message}\n'
    raise MagnetronError(msg, info)



from __future__ import annotations

from magnetron import nn, optim, Tensor
from matplotlib import pyplot as plt

EPOCHS: int = 2000
LR: float = 1e-1

# Define the XOR input and output data
x = Tensor.of([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
y = Tensor.of([[0.0], [1.0], [1.0], [0.0]])

# Create the model, loss function, and optimizer
model = nn.Sequential(nn.Linear(2, 2), nn.Tanh(), nn.Linear(2, 1), nn.Tanh())
optimizer = optim.SGD(model.parameters(), lr=LR)
criterion = nn.MSELoss()

# Train the model
losses: list[float] = []
for epoch in range(EPOCHS):
    y_hat = model(x)
    loss = criterion(y_hat, y)
    loss.backward()
    optimizer.step()
    optimizer.zero_grad()
    losses.append(loss.item())
    if epoch % 100 == 0:
        print(f'Epoch: {epoch}, Loss: {loss.item()}')

# Print predictions
y_hat = model(x)
for i in range(x.shape[0]):
    print(f'Expected: {y[i].item()}, Predicted: {y_hat[i].item()}')

# Plot results
plt.figure()
plt.plot(losses)
plt.xlabel('Epoch')
plt.ylabel('MSE Loss')
plt.title('Training Loss over Time')
plt.grid(True)
plt.show()



# (c) 2025 Mario Sieg. <mario.sieg.64@gmail.com>

from magnetron import context


def test_context_creation() -> None:
    # Test that a context can be created and defaults are correct.
    ctx = context.native_ptr()
    ctx = context.native_ptr()
    ctx = context.native_ptr()
    ctx = context.native_ptr()
    assert isinstance(context.os_name(), str)
    assert isinstance(context.cpu_name(), str)
    assert context.cpu_virtual_cores() >= 1
    assert context.cpu_physical_cores() >= 1



# (c) 2025 Mario 'Neo' Sieg. <mario.sieg.64@gmail.com>

from ..common import *


def _maybe_negative_axes(axes, nd):
    out = []
    for ax in axes:
        out.append(ax - nd if random.random() < 0.5 else ax)
    return out


def reduce_op_with_random_axes(
        dtype: DataType,
        make_mag: Callable[[list[int] | None, bool], Callable[[Tensor], Tensor]],
        make_torch: Callable[[list[int] | None, bool], Callable[[torch.Tensor], torch.Tensor]],
        keepdim: bool,
) -> None:
    def test(shape: tuple[int, ...]) -> None:
        nd = len(shape)
        k = random.randint(0, nd)
        axes = sorted(random.sample(range(nd), k))
        axes = _maybe_negative_axes(axes, nd)
        dim_arg = axes if axes else None

        x = Tensor.bernoulli(shape) if dtype == boolean else Tensor.uniform(shape, dtype=dtype)
        mag_cb = make_mag(dim_arg, keepdim)
        torch_cb = make_torch(dim_arg, keepdim)

        r = mag_cb(x.clone())
        r_t = totorch(r)
        t = torch_cb(totorch(x))

        # fix scalar vs (1,) mismatch (see next section)
        if r_t.ndim == 1 and r_t.shape[0] == 1 and t.ndim == 0:
            t = t.reshape(1)

        torch.testing.assert_close(r_t, t)

    for_all_shapes(test)


@pytest.mark.parametrize('dtype', dtype.FLOATING_POINT_DTYPES)
@pytest.mark.parametrize('keepdim', [False, True])
def test_reduction_mean(dtype: DataType, keepdim: bool):
    reduce_op_with_random_axes(
        dtype,
        make_mag   = lambda dim, keep: (lambda x: x.mean(dim=dim, keepdim=keep)),
        make_torch = lambda dim, keep: (lambda x: x.mean(dim=dim, keepdim=keep)),
        keepdim=keepdim,
    )

@pytest.mark.parametrize('dtype', dtype.NUMERIC_DTYPES)
@pytest.mark.parametrize('keepdim', [False, True])
def test_reduction_sum(dtype: DataType, keepdim: bool):
    reduce_op_with_random_axes(
        dtype,
        make_mag   = lambda dim, keep: (lambda x: x.sum(dim=dim, keepdim=keep)),
        make_torch = lambda dim, keep: (lambda x: x.sum(dim=dim, keepdim=keep)),
        keepdim=keepdim,
    )

@pytest.mark.parametrize('dtype', dtype.NUMERIC_DTYPES)
@pytest.mark.parametrize('keepdim', [False, True])
def test_reduction_prod(dtype: DataType, keepdim: bool):
    reduce_op_with_random_axes(
        dtype,
        make_mag   = lambda dim, keep: (lambda x: x.prod(dim=dim, keepdim=keep)),
        make_torch = lambda dim, keep: (lambda x: x.prod(dim=dim, keepdim=keep)),
        keepdim=keepdim,
    )

@pytest.mark.parametrize('dtype', dtype.NUMERIC_DTYPES)
@pytest.mark.parametrize('keepdim', [False, True])
def test_reduction_min(dtype: DataType, keepdim: bool):
    reduce_op_with_random_axes(
        dtype,
        make_mag   = lambda dim, keep: (lambda x: x.min(dim=dim, keepdim=keep)),
        make_torch = lambda dim, keep: (lambda x: x.amin(dim=dim, keepdim=keep)),
        keepdim=keepdim,
    )

@pytest.mark.parametrize('dtype', dtype.NUMERIC_DTYPES)
@pytest.mark.parametrize('keepdim', [False, True])
def test_reduction_max(dtype: DataType, keepdim: bool):
    reduce_op_with_random_axes(
        dtype,
        make_mag   = lambda dim, keep: (lambda x: x.max(dim=dim, keepdim=keep)),
        make_torch = lambda dim, keep: (lambda x: x.amax(dim=dim, keepdim=keep)),
        keepdim=keepdim,
    )



from __future__ import annotations

import argparse

from magnetron import nn, optim, context, Tensor, no_grad, dtype
import matplotlib.pyplot as plt


class AE(nn.Module):
    def __init__(self, w: int, h: int, latent_dim: int = 16) -> None:
        super().__init__()
        self.latent_dim = latent_dim
        self.w = w
        self.h = h
        self.encoder = nn.Sequential(
            nn.Flatten(),
            nn.Linear(3 * w * h, latent_dim),
            nn.ReLU(),
        )
        self.decoder = nn.Sequential(
            nn.Linear(latent_dim, 3 * w * h),
            nn.Sigmoid(),
        )

    def forward(self, x: Tensor) -> Tensor:
        y2 = self.decoder(self.encoder(x))
        return y2.view(x.shape[0], 3, self.w, self.h)


def _main() -> None:
    args = argparse.ArgumentParser(description='Autoencoder Example')
    args.add_argument('--image', type=str, default='media/logo.png', help='Path to input image')
    args.add_argument('--epochs', type=int, default=100, help='Number of training epochs')
    args.add_argument('--latent', type=int, default=16, help='Dimensionality of the latent space')
    args.add_argument('--width', type=int, default=64, help='Resized image width')
    args.add_argument('--height', type=int, default=64, help='Resized image height')
    args.add_argument('--seed', type=int, default=3407, help='Random seed for reproducibility')
    args.add_argument('--lr', type=float, default=1e-3, help='Learning rate for the optimizer')

    args = args.parse_args()
    print(args)

    context.manual_seed(args.seed)

    # Load and preprocess image
    image = Tensor.load_image(args.image, channels='RGB', resize_to=(args.width, args.height))  # Load image into uint8 CxHxW tensor
    image = (image.cast(dtype.float32) / 255)[None, ...]  # Convert uint8 -> float tensor and normalize [0, 255) to [0, 1) and insert batch dim

    # Initialize model, loss function, and optimizer
    model = AE(w=args.width, h=args.height, latent_dim=args.latent)
    criterion = nn.MSELoss()
    optimizer = optim.Adam(model.parameters(), lr=args.lr)

    # Train the autoencoder
    model.train()
    losses: list[float] = []

    for step in range(args.epochs):
        y_hat = model(image)
        loss = criterion(y_hat, image)
        loss.backward()
        optimizer.step()
        optimizer.zero_grad()
        losses.append(loss.item())
        if step % 10 == 0:
            print(f'Epoch [{step}/{args.epochs}] Loss: {loss.item():.6f}')

    print('Training complete')

    # Reconstruct image
    model.eval()
    with no_grad():
        reconstructed = model(image)

    # Plot original and reconstructed images
    original = image[0].permute(1, 2, 0).tolist()
    reconstructed = reconstructed[0].permute((1, 2, 0)).tolist()
    plt.figure()
    plt.subplot(1, 2, 1)
    plt.title('Original')
    plt.imshow(original)
    plt.axis('off')
    plt.subplot(1, 2, 2)
    plt.title('Reconstructed')
    plt.imshow(reconstructed)
    plt.axis('off')
    plt.tight_layout()
    # plt.savefig('autoencoder_result.png', dpi=300)

    plt.figure()
    plt.plot(losses)
    plt.xlabel('Epoch')
    plt.ylabel('MSE Loss')
    plt.title('Training Loss over Time')
    plt.grid(True)
    plt.tight_layout()
    # plt.savefig('autoencoder_loss.png', dpi=300)
    plt.show()


if __name__ == '__main__':
    _main()



# (c) 2025 Mario 'Neo' Sieg. <mario.sieg.64@gmail.com>

from __future__ import annotations

import itertools
import multiprocessing
import random
from collections.abc import Callable
from typing import Any, Iterator

import torch
import pytest
from magnetron import *
from collections import deque
from enum import Enum, unique

# Give torch 1//4 of the total cores to not overload the CPU with inner parallelism threads + parallel spawned pytests
torch.set_num_threads(max(4, multiprocessing.cpu_count() // 8))
torch.set_num_interop_threads(max(4, multiprocessing.cpu_count() // 8))

DTYPE_TORCH_MAP: dict[DataType, torch.dtype] = {
    float16: torch.float16,
    float32: torch.float32,
    boolean: torch.bool,
    uint8: torch.uint8,
    int8: torch.int8,
    uint16: torch.uint16,
    int16: torch.int16,
    uint32: torch.uint32,
    int32: torch.int32,
    uint64: torch.uint64,
    int64: torch.int64,
}


def totorch_dtype(dtype: DataType) -> torch.dtype:
    if dtype not in DTYPE_TORCH_MAP:
        raise ValueError(f'Unsupported dtype: {dtype}')
    return DTYPE_TORCH_MAP[dtype]


def totorch(obj: Tensor | int | float | bool, dtype: torch.dtype | None = None) -> torch.Tensor:
    if isinstance(obj, torch.Tensor):
        return obj.to(dtype) if dtype is not None else obj
    if not isinstance(obj, Tensor):
        return torch.tensor(obj, dtype=dtype) if dtype is not None else torch.tensor(obj)
    if dtype is None:
        dtype = totorch_dtype(obj.dtype)
    t = torch.tensor(obj.tolist(), dtype=dtype)
    if tuple(obj.shape) == (1,):
        return t.reshape(())
    return t.reshape(obj.shape)


def broadcastable(a: tuple[int, ...], b: tuple[int, ...]) -> bool:
    for x, y in zip(a[::-1], b[::-1]):
        if not (x == y or x == 1 or y == 1):
            return False
    return True


def broadcast_shape(a: tuple[int, ...], b: tuple[int, ...]) -> tuple[int, ...]:
    rev = []
    for x, y in zip(a[::-1], b[::-1]):
        rev.append(max(x, y))
    longer = a if len(a) > len(b) else b
    rev.extend(longer[: abs(len(a) - len(b))][::-1])
    return tuple(rev[::-1])


def iter_shapes(rank: int, lim: int) -> Iterator[tuple[int, ...]]:
    if rank == 0:
        yield ()
        return
    tup = [1] * rank
    while True:
        yield tuple(tup)
        i = rank - 1
        while i >= 0 and tup[i] == lim:
            tup[i] = 1
            i -= 1
        if i < 0:
            break
        tup[i] += 1


def matmul_shape_pairs(lim: int, max_total_rank: int = 6) -> Iterator[tuple[tuple[int, ...], tuple[int, ...]]]:
    max_batch_rank = max_total_rank - 2
    rng = range(1, lim + 1)
    for batch_rank in range(max_batch_rank + 1):
        for batch_a in iter_shapes(batch_rank, lim):
            for batch_b in iter_shapes(batch_rank, lim):
                if not broadcastable(batch_a, batch_b):
                    continue
                batched = broadcast_shape(batch_a, batch_b)
                for M in rng:
                    for K in rng:
                        for N in rng:
                            shape_A = (*batched, M, K)
                            shape_B = (*batched, K, N)
                            yield shape_A, shape_B


TEST_SHAPES: tuple[tuple[int], ...] = (
    (1,),
    (1, 1),
    (1, 1, 1),
    (1, 1, 1, 1),
    (1, 1, 1, 1, 1, 1, 1, 1),
    (2,),
    (3,),
    (4,),
    (1, 2),
    (2, 1),
    (1, 3),
    (3, 1),
    (1, 1, 2),
    (1, 2, 1),
    (2, 1, 1),
    (2, 3),
    (3, 2),
    (1, 2, 3),
    (3, 2, 1),
    (1, 3, 1, 3),
    (3, 5),
    (4, 7),
    (7, 4),
    (5, 3, 4),
    (4, 5, 3),
    (3, 4, 5),
    (1, 16),
    (16, 1),
    (1, 1, 16),
    (8, 1, 16),
    (1, 8, 16),
    (2, 3, 1, 16),
    (2, 1, 3, 16),
    (1, 2, 3, 1),
    (1, 2, 1, 3),
    (2, 1, 1, 3),
    (1, 32, 128),
    (32, 1, 128),
    (32, 128, 1),
    (16,),
    (32,),
    (64,),
    (128,),
    (256,),
    (512,),
    (1024,),
    (2048,),
    (4096,),
    (64, 64),
    (128, 128),
    (256, 256),
    (512, 512),
    (1024, 1024),
    (1, 3, 224, 224),
    (1, 3, 512, 512),
    (8, 3, 32, 32),
    (4, 3, 1024, 1024),
    (64, 3, 7, 7),
    (128, 64, 3, 3),
    (256, 128, 1, 1),
    (512, 256, 3, 3),
    (1, 197, 768),
    (1, 577, 768),
    (1, 197, 1024),
    (1, 4096),
    (1, 5120),
    (1, 8192),
    (1, 12288),
    (16, 4096),
    (32, 4096),
    (8, 8192),
    (8, 12288),
    (32, 128),
    (40, 128),
    (64, 128),
    (32, 96),
    (4096 // 4, 12288 // 4),
    (8192 // 4, 8192 // 4),
    (12288 // 4, 12288 // 4),
    (11008 // 4, 4096 // 4),
    (4096 // 4, 11008 // 4),
    (8192 // 4, 22016 // 6),
    (12288 // 4, 49152 // 8),
    (1, 32, 2048 // 4, 128 // 4),
    (4, 32, 2048 // 4, 128),
    (1, 40, 4096 // 4, 96),
    (8, 64, 1024 // 4, 128 // 4),
    (2, 32, 4096 // 4, 128),
    (16, 128),
    (128, 128),
    (2048, 128),
    (3, 1, 7),
    (1, 3, 7),
    (7, 1, 3),
    (3, 7, 1),
    (1, 7, 3),
    (7, 3, 1),
    (2, 3, 5, 7),
    (7, 5, 3, 2),
    (17,),
    (31,),
    (67,),
    (127,),
    (257,),
    (2, 3, 4, 5, 6),
    (6, 5, 4, 3, 2),
    (1, 2, 3, 4, 5, 6),
    (1, 1, 4096, 4096),
    (1024 // 2, 1024 // 2),
    (2048 // 2, 2048 // 2),
    (4096 // 2, 1024 // 2),
    (1024 // 2, 4096 // 2),
    (4096 // 2, 4096 // 2),
    (8192 // 2, 2048 // 2),
    (6, 66, 666),
    (4, 4, 1024, 1024),
    (2, 8, 512, 512),
    (5,),
    (7,),
    (9,),
    (11,),
    (13,),
    (2, 2),
    (2, 2, 2),
    (2, 2, 2, 2),
    (3, 3),
    (3, 3, 3),
    (4, 1, 4),
    (1, 4, 1, 4),
    (1, 2, 2, 1),
    (2, 1, 2, 1),
    (1, 1, 2, 2),
    (2, 2, 1, 1),
    (1, 3, 5),
    (5, 3, 1),
    (1, 5, 3),
    (3, 1, 5),
    (1, 1, 8),
    (8, 1, 1),
    (1, 8, 1),
    (1, 1, 1, 8),
    (8, 1, 1, 1),
    (1, 8, 1, 1),
    (1, 1, 8, 1),
    (1, 1, 16, 1),
    (1, 16, 1, 1),
    (1, 4, 1, 16),
    (4, 1, 16, 1),
    (1, 1, 4, 16),
    (1, 7, 1, 1, 7),
    (7, 1, 7, 1, 1),
    (1, 1, 7, 1, 7),
    (1, 3, 7, 7),
    (1, 3, 9, 9),
    (1, 3, 11, 11),
    (2, 3, 17, 17),
    (1, 1, 15, 15),
    (4, 3, 13, 13),
    (8, 3, 17, 17),
    (1, 16, 64),
    (2, 16, 64),
    (4, 16, 64),
    (1, 32, 64),
    (2, 32, 64),
    (4, 32, 64),
    (1, 32, 128),
    (2, 32, 128),
    (4, 32, 128),
    (1, 64, 64),
    (2, 64, 64),
    (4, 64, 64),
    (1, 64, 128),
    (2, 64, 128),
    (4, 64, 128),
    (1, 1, 16, 16),
    (1, 2, 16, 32),
    (2, 2, 16, 32),
    (1, 4, 32, 32),
    (2, 4, 32, 32),
    (4, 4, 32, 32),
    (1, 8, 64, 32),
    (2, 8, 64, 32),
    (1, 8, 32, 64),
    (2, 8, 32, 64),
    (7, 13),
    (13, 7),
    (15, 33),
    (33, 15),
    (31, 64),
    (64, 31),
    (63, 64),
    (64, 63),
    (127, 64),
    (64, 127),
    (255, 64),
    (64, 255),
    (16, 20),
    (20, 16),
    (8, 24),
    (24, 8),
    (7, 17),
    (17, 7),
    (5, 21),
    (21, 5),
    (4, 4, 16),
    (4, 8, 16),
    (8, 4, 16),
    (2, 8, 32),
    (2, 16, 32),
    (4, 8, 32),
    (2, 4, 8, 16),
    (2, 4, 16, 8),
    (4, 2, 8, 16),
    (1, 4, 8, 16),
    (1, 8, 4, 16),
    (1, 2, 3, 4, 5),
    (5, 4, 3, 2, 1),
    (1, 1, 2, 3, 4),
    (1, 2, 1, 3, 4),
    (2, 1, 3, 1, 4),
    (1, 2, 3, 1, 4),
    (1, 2, 3, 4, 1),
    (1, 2, 3, 4, 5, 1),
    (1, 1, 2, 3, 4, 5),
    (2, 3, 1, 4, 1, 5),
    (1, 16, 8, 8),
    (1, 8, 8, 16),
    (2, 16, 8, 8),
    (2, 8, 8, 16),
    (4, 16, 4, 4),
    (4, 4, 4, 16),
    (1, 1, 32),
    (32, 1, 1),
    (1, 32, 1),
    (2, 1, 32),
    (2, 32, 1),
    (1, 2, 32),
    (1, 2, 1, 32),
    (2, 1, 1, 32),
    (1, 2, 32, 1),
    (1, 15),
    (1, 31),
    (1, 63),
    (2, 63),
    (4, 63),
    (3, 21),
    (3, 63),
    (5, 19),
    (7, 8),
    (8, 7),
    (15, 16),
    (16, 15),
    (31, 32),
    (32, 31),
    (1, 2, 2, 2, 2, 2, 2),
    (2, 1, 2, 2, 2, 2, 2),
    (1, 3, 1, 3, 1, 3, 1),
    (2, 3, 4, 1, 2, 3, 4),
    (1, 2, 3, 4, 1, 2, 3, 4),
)


def for_all_shapes(f: Callable[tuple[int, ...]]) -> None:
    for shape in TEST_SHAPES:
        f(shape)


@unique
class BinaryOpParamKind(Enum):
    TENSOR = 'tensor'
    SCALAR = 'scalar'
    LIST = 'list'


def _allocate_binary_op_args(
    dtype: DataType, shape: tuple[int, ...], kind: BinaryOpParamKind, low: float | int = 0, high: float | int = 1
) -> tuple[Tensor, Tensor | list[Any] | float | int]:
    if dtype == boolean:
        x = Tensor.bernoulli(shape)
        match kind:
            case BinaryOpParamKind.TENSOR:
                return x, Tensor.bernoulli(shape)
            case BinaryOpParamKind.LIST:
                return x, [random.choice([True, False]) for _ in range(nested_len(list(shape)))]
            case BinaryOpParamKind.SCALAR:
                return x, random.choice([True, False])
            case _:
                raise ValueError(f'Unknown BinaryOpParamKind: {kind}')
    else:
        x = Tensor.uniform(shape, dtype=dtype, low=low, high=high)
        match kind:
            case BinaryOpParamKind.TENSOR:
                return x, Tensor.uniform(shape, dtype=dtype, low=low, high=high)
            case BinaryOpParamKind.LIST:
                if dtype.is_integer:
                    return x, [random.randint(low, high) for _ in range(nested_len(list(shape)))]
                else:
                    return x, [random.uniform(low, high) for _ in range(nested_len(list(shape)))]
            case BinaryOpParamKind.SCALAR:
                if dtype.is_integer:
                    return x, random.randint(low, high)
                else:
                    return x, random.uniform(low, high)
            case _:
                raise ValueError(f'Unknown BinaryOpParamKind: {kind}')


def binary_op_square(
    dtype: DataType,
    callback: Callable[[Tensor | torch.Tensor, Tensor | torch.Tensor], Tensor | torch.Tensor],
    kind: BinaryOpParamKind = BinaryOpParamKind.TENSOR,
    low: float | int = 0,
    high: float | int = 1,
) -> None:
    def func(shape: tuple[int, ...]) -> None:
        x, y = _allocate_binary_op_args(dtype, shape, kind, low, high)
        r = callback(x, y)
        torch.testing.assert_close(totorch(r), callback(totorch(x), totorch(y)))

    for_all_shapes(func)


def binary_cmp_op(
    dtype: DataType,
    callback: Callable[[Tensor | torch.Tensor, Tensor | torch.Tensor], Tensor | torch.Tensor],
    kind: BinaryOpParamKind = BinaryOpParamKind.TENSOR,
    low: float | int = 0,
    high: float | int = 1,
) -> None:
    def func(shape: tuple[int, ...]) -> None:
        x, y = _allocate_binary_op_args(dtype, shape, kind, low, high)
        r = callback(x, y)
        assert r.dtype == boolean
        torch.testing.assert_close(totorch(r, torch.bool), callback(totorch(x), totorch(y)))

    for_all_shapes(func)


def scalar_op(dtype: DataType, callback: Callable[[Tensor | torch.Tensor, int | float | bool], Tensor | torch.Tensor], rhs: bool = True) -> None:
    def func(shape: tuple[int, ...]) -> None:  # x op scalar
        xi: float = random.uniform(-1.0, 1.0)
        x = Tensor.uniform(shape, dtype=dtype)
        r = callback(x, xi)
        torch.testing.assert_close(totorch(r), callback(totorch(x), xi))

    for_all_shapes(func)

    if not rhs:
        return

    def func(shape: tuple[int, ...]) -> None:  # scalar op x
        xi: float = random.uniform(-1.0, 1.0)
        x = Tensor.uniform(shape)
        r = callback(xi, x)
        torch.testing.assert_close(totorch(r), callback(xi, totorch(x)))

    for_all_shapes(func)


def nested_len(obj: list[Any]) -> int:
    total = 0
    stack = deque([obj])
    seen = set()
    while stack:
        current = stack.pop()
        obj_id = id(current)
        if obj_id in seen:
            continue
        seen.add(obj_id)
        for item in current:
            if isinstance(item, list):
                stack.append(item)
            else:
                total += 1
    return total


def flatten(nested: Any) -> list[Any]:
    out: list[Any] = []
    stack = deque([iter(nested)])
    while stack:
        try:
            item = next(stack[-1])
        except StopIteration:
            stack.pop()
            continue
        if isinstance(item, list) or isinstance(item, tuple):
            stack.append(iter(item))
        else:
            out.append(item)
    return out
